/**
 * Copyright (c) 2005-2012 EBM WebSourcing, 2012-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.api.communication.sharedarea;

import java.util.Collection;
import java.util.Properties;

import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;

import org.ow2.petals.jbi.servicedesc.endpoint.PetalsServiceEndpoint;
import org.ow2.petals.jbi.servicedesc.endpoint.PetalsServiceEndpoint.EndpointType;
import org.ow2.petals.microkernel.api.communication.sharedarea.exception.InvalidPropertyException;
import org.ow2.petals.microkernel.api.communication.sharedarea.exception.MissingPropertyException;
import org.ow2.petals.microkernel.api.communication.sharedarea.exception.SharedAreaException;
import org.ow2.petals.microkernel.api.configuration.ConfigurationService;
import org.ow2.petals.microkernel.api.implementation.MutableImplementation;
import org.ow2.petals.microkernel.api.implementation.exception.MutableImplException;
import org.ow2.petals.microkernel.api.server.FractalHelper;
import org.w3c.dom.Document;

import com.google.common.base.Optional;

/**
 * Interface of the component responsible of the shared area management. The shared area is used a distributed shared
 * memory to store information shared by all containers of a same domain.
 * 
 * It is responsible of actually adding the container to its current topology in the shared memory on component start.
 * But in order to leave a topology, one must call {@link #leaveTopology()}: when the component is stopped, the
 * container is not removed from the topology in the shared memory.
 * 
 * The idea is that adding or removing a container to a topology is a administration operation.
 * 
 * All the other operations are related to endpoint management: the topology operations are reduced to their minimum in
 * this interface.
 * 
 * @author ddesjardins - EBM WebSourcing
 * @author Olivier Fabre - EBM WebSourcing
 * @author Christophe DENEUX - Linagora
 * @author Victor Noel - Linagora
 */
public interface SharedAreaService extends MutableImplementation {

    public static final String COMPONENT_LOGGER_NAME = "Petals.Communication.SharedAreaService";

    /**
     * Name of the server interface of the Fractal component {@value FractalHelper#SHARED_AREA_COMPONENT}
     */
    public static final String FRACTAL_SRV_ITF_NAME = "service";

    /**
     * The name of the client interface of the component {@value FractalHelper#SHARED_AREA_COMPONENT} for the
     * configuration service.
     */
    public final static String CONFIGURATION_FRACTAL_ITF_NAME = "configuration";

    /**
     * <p>
     * Retrieve the configuration of the mutable implementation as {@link Properties}. See API of the mutable
     * implementation for the property list.
     * </p>
     * <p>
     * All properties are prefixed by the class name of the mutable implementation
     * </p>
     * 
     * @param providedTopologyPassphrase
     *            Passphrase needed to return sensible information. If <code>null</code>, empty or not matching the one
     *            of the container on which the connection is established, no sensible information will be returned.
     * @return The configuration of the mutable implementation as {@link Properties}. Not <code>null</code>.
     * @throws MutableImplException
     *             An error occurs getting the configuration of the mutable implementation
     */
    public Properties getConfiguration(final String providedTopologyPassphrase) throws MutableImplException;

    /**
     * Get the XML part, located into the topology definition, of the mutable implementation configuration from a
     * property list generated by {@link #getConfiguration(String)}
     * 
     * @param properties
     *            A property list
     * @return The XML part as {@link Document}, located into the topology definition, of the mutable implementation
     *         configuration, or <code>null</code> if:
     *         <ul>
     *         <li>the given parameter <code>properties</code> is null</li>
     *         <li>no configuration is required by the mutable implementation</li>
     *         </ul>
     * @throws MissingPropertyException
     *             A required property is missing.
     * @throws InvalidPropertyException
     *             A property has an invalid value.
     * @throws MutableImplException
     *             An error occurs creating the DOM {@link Document} associated to the topology part of the
     *             configuration
     */
    public Document getTopologyPartConfiguration(final Properties properties) throws MutableImplException;

    /**
     * <p>
     * Register an internal endpoint in the shared area.
     * </p>
     * 
     * @param endpointName
     *            Endpoint name
     * @param serviceName
     *            Endpoint service name.
     * @param interfaces
     *            Endpoint interfaces
     * @param componentName
     *            the name of the component activating this endpoint. Not <code>null</code>.
     * @param description
     *            Endpoint WSDL
     * @return A {@link PetalsServiceEndpoint} for the registered endpoint or <code>null</code> if the endpoint already
     *         exists.
     * @throws SharedAreaException
     *             An error occurs registering the endpoint
     */
    public PetalsServiceEndpoint registerInternalEndpoint(String endpointName, QName serviceName, QName[] interfaces,
            String componentName, Document description) throws SharedAreaException;

    /**
     * <p>
     * Register an endpoint in the shared area.
     * </p>
     * 
     * @param endpoint
     *            Endpoint
     * @param componentName
     *            the name of the component activating this endpoint. Not <code>null</code>.
     * @return A {@link PetalsServiceEndpoint} for the registered endpoint or <code>null</code> if the endpoint already
     *         exists.
     * @throws SharedAreaException
     *             An error occurs registering the endpoint
     */
    public PetalsServiceEndpoint registerExternalEndpoint(ServiceEndpoint endpoint, String componentName)
            throws SharedAreaException;

    /**
     * Deregister the given endpoint.
     * 
     * @param serviceName
     *            qualified name of the endpoint's service; must be non null
     * @param endpointName
     *            name of the endpoint; must be non null
     * @param componentName
     *            the name of the component activating this endpoint; must be non null
     * @param type
     *            The type that the endpoints must implement. must be non null
     * @return the removed endpoint
     * @throws SharedAreaException
     *             An error occurred deregistering the endpoint
     */
    public PetalsServiceEndpoint deregisterEndpoint(QName serviceName, String endpointName, String componentName,
            EndpointType type) throws SharedAreaException;

    /**
     * <p>
     * Retrieve registered endpoints. The returned collection is not backed by the shared area.
     * </p>
     * 
     * @param type
     *            The type that the endpoints must implement. Can be <code>null</code>.
     * @return The endpoints matching '<code>type</code>' or all endpoints if '<code>type</code>' is <code>null</code>
     */
    public Collection<PetalsServiceEndpoint> getEndpoints(EndpointType type) throws SharedAreaException;

    /**
     * <p>
     * Get the service endpoint for the named endpoint, if any (note that type can't be null because we can return only
     * one endpoint).
     * </p>
     * 
     * <p>
     * Note: this method could be implemented at endpoint directory level getting all endpoints and filtering them, but
     * for performance reasons some optimization can be done if the shared area provides directly such a method.
     * </p>
     * 
     * 
     * @param serviceName
     *            qualified name of the endpoint's service; must be non null
     * @param endpointName
     *            name of the endpoint; must be non null
     * @param type
     *            The type that the endpoints must implement. must be non null.
     * @return the named endpoint, or null if the named endpoint is not registered
     */
    public PetalsServiceEndpoint getEndpoint(QName serviceName, String endpointName, EndpointType type)
            throws SharedAreaException;

    /**
     * <p>
     * Get endpoints implementing the interface '<code>interfaceName</code>'. The returned collection is not backed by
     * the shared area.
     * </p>
     * 
     * @param interfaceName
     *            qualified name of interface implemented by the endpoints; must be non-null
     * @param type
     *            The type that the endpoints must implement. Can be <code>null</code>.
     * @return the registered endpoints for the specified interface name;
     * @throws SharedAreaException
     *             A parameter is invalid, or an error occurs searching endpoints
     */
    public Collection<PetalsServiceEndpoint> getEndpointsForInterface(QName interfaceName, EndpointType type)
            throws SharedAreaException;

    /**
     * <p>
     * Get endpoints implementing the service '<code>serviceName</code>'. The returned collection is not backed by the
     * shared area.
     * </p>
     * 
     * @param serviceName
     *            The service that the endpoints must implement. Must be non-null.
     * @param type
     *            The type that the endpoints must implement. Can be <code>null</code>.
     * @return The endpoints matching '<code>serviceName</code>'
     * @throws SharedAreaException
     *             A parameter is invalid, or an error occurs searching endpoints
     */
    public Collection<PetalsServiceEndpoint> getEndpointsForService(QName serviceName, EndpointType type)
            throws SharedAreaException;

    /**
     * Returns the ServiceEndpoint description as a DOM Document
     * 
     * @param endpoint
     *            the endpoint
     * @return a DOM representation of a ServiceEndpoint description or <code>null</code> if the endpoint is not a valid
     *         reference
     * @throws SharedAreaException
     *             if an error occurs during endpoint description retrieval
     */
    public Optional<Document> getDescription(ServiceEndpoint endpoint) throws SharedAreaException;

    /**
     * Returns an internal endpoint description as a DOM Document
     * 
     * @param serviceName
     *            the service name
     * @param endpointName
     *            the endpoint name
     * @return a {@link String} representation of a ServiceEndpoint description or <code>null</code> if there is no such
     *         description
     * @throws SharedAreaException
     *             if an error occurs during endpoint description retrieval
     */
    public String getDescription(QName serviceName, String endpointName) throws SharedAreaException;

    /**
     * Leave the current topology (if it's a dynamic topology, we remove ourselves from it) and stops listening to
     * topology changes.
     * 
     * It also removes all the local endpoints and returns them to be restored using
     * {@link #restoreEndpoints(Collection)}.
     * 
     * @return the list of removed endpoints
     * @throws SharedAreaException
     */
    public Collection<SavedEndpoint> leaveTopology() throws SharedAreaException;

    /**
     * Join the current topology configured in the {@link ConfigurationService} (stored in the shared area or if none
     * present, locally in topology.xml) and start listening to topology changes.
     * 
     * If the <code>addMyself</code> argument is <code>true</code> and it's a dynamic topology, we add ourselves to it.
     * 
     * It also marks endpoints already associated to the present container as being in restoration. To conclude this
     * process, {@link #restoreEndpoints(Collection)} should be called after.
     * 
     * @throws SharedAreaException
     *             if we can't add ourselves (</code>addMyself</code> being <code>true</code> and topology is static) or
     *             if we are not already in it (</code>addMyself</code> being <code>false</code>).
     */
    public void joinTopology(final boolean addMyself) throws SharedAreaException;

    /**
     * Restore all the local endpoints passed as argument. It will also remove the endpoints in the shared area that
     * were bound to this container in the topology during the attach/connect and that were not restored.
     * 
     * This method must be called after {@link #joinTopology(boolean)}, optionally activating endpoints (e.g. during
     * recovery) inbetween.
     * 
     */
    public void restoreEndpoints(Collection<SavedEndpoint> endpoints) throws SharedAreaException;

    public class SavedEndpoint {

        public final Document description;

        public final PetalsServiceEndpoint endpoint;

        public SavedEndpoint(final PetalsServiceEndpoint endpoint, final Document description) {
            this.endpoint = endpoint;
            this.description = description;
        }
    }
}
